/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2008 The LS-Universe Team

  Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#if !defined(__LABEL_H)
#define __LABEL_H

#include "GhostLabel.h"

class Label_Region
{
public:

	Label_Region() : 
		strOrgX("0"),
		strOrgY("0"),
		strOrgWidth("0"),
		strOrgHeight("0"),

		strOriginalTooltip(""),
		strTooltip(""),
		strAction(""),

		pGhostLabel(NULL)
	{
		region.left = 0;
		region.right = 0;
		region.top = 0;
		region.bottom = 0;
	};

	virtual ~Label_Region() {};

	string strOrgX;
	string strOrgY;
	string strOrgWidth;
	string strOrgHeight;
	
	RECT region;

	string strOriginalTooltip;
	string strTooltip;
	string strAction;

	GhostLabel *pGhostLabel;
};

typedef vector<GhostLabel *>GhostLabelList;
typedef vector<GhostLabel *>::iterator GhostLabelListIterator;

typedef vector<Label_Region *>Label_RegionList;
typedef vector<Label_Region *>::iterator Label_RegionListIterator;

struct ModClick
{
	int iModifier;
	string strAction;
};

typedef vector<ModClick *>ModClickList;
typedef vector<ModClick *>::iterator ModClickListIterator;

class Label
{
public:

	Label(const string &strLabelName);
	virtual ~Label();

	void load(HINSTANCE hInstance, HWND box = NULL, bool bDropShadow = false);
	void reconfigure(bool bRefresh = false);

	void CreateOverlay(string strConfig, bool bUpdateDimensions = false);
	void ModifyOverlay(string strOverlayName, string strConfig);
	void DestroyOverlay(string strOverlayName);

	bool isAlwaysOnTop() const { return bAlwaysOnTop; };
	bool isNormalWindowzOrder() const { return bNormalWindowzOrder; };
	bool isVisible() const { return bVisible; };
	bool isTextboxVisible() const { return bTextboxVisible; };
	bool isStartHidden() const { return bStartHidden; };
	bool isGhosted() const { return bGhosted; };

	//int getX() const { return actualx; };
	//int getY() const { return actualy; };
	//int getHeight() const { return actualheight; };
	//int getWidth() const { return actualwidth; };

	int getTrueTransparency() const { return bTrueTransparency; };
	int getScrolling() const { return iScroll; };

	const HWND &getBox() const { return box; };
	const HWND &gethWnd() const { return hWnd; };
	const HWND &getTexteditBoxhWnd() const { return hWndTextBox; };
	const string &getName() const { return strName; };
	const string &getLabelGroup() const { return strLabelGroup; };
	const string &getText() const { return text; };

	HCURSOR getTextCursor() { return textcursor; };
	string getTextCursorPath() { return textcursorPath; };

	bool hasCorrectAlphaState() { return ((!alphaMap && !fakealphaMap) ? true : false); };
	void setFakeAlphaMap(bool newState) { fakealphaMap = newState; };

	void setAlwaysOnTop(bool bNewAlwaysOnTop, bool bFirstTime = false);
	void setGhosted(bool setghosted);
	void setScrolling(int scrolling);
	void setScrollLimit(int limit);
	
	void setAlpha(int newAlphaTransparency, bool hookingReset = false);
	void setAlphaFade(int newAlphaTransparency, int step, int delay);

	void setAnimation(bool start, int loops = -1);

	bool checkRegions(int x, int y, Label_RegionList inputlist, bool btooltip = true);
	bool checkClickRegions(int x, int y, Label_RegionList inputlist, int checkPressedState = 0, int actualMouseClick = -1);
	
	bool checkHREFRegions(int x, int y, vector<string> inputlist);
	bool checkHREFClickRegions(int x, int y, vector<string> inputlist);
	
	void repaint(bool bDeleteScrollBackup = false);
	void updateRegion();

	void reposition(int type, int newx, int newy, int newwidth, int newheight, int steps = 0, int time = 0, bool repositionBy = false, bool movecommand = true, string strSender = "");

	//ExtraFeatureMode = 0 Normal Label
	//ExtraFeatureMode = 1 TextEditBox
	//ExtraFeatureMode = 2 GhostLabel
	void hide(bool execute = false, int extraFeatureMode = 0, GhostLabel *ghostlabel = NULL);
	void show(bool execute = false, int extraFeatureMode = 0, GhostLabel *ghostlabel = NULL);
	void showHide(int timeout, bool execute = false);
	
	void setText(const string &text, bool updateBackup = false, int extraFeatureMode = 0, GhostLabel *ghostlabel = NULL);

	void previous();
	void next();
	void update(bool bForceRepaint = false);

	void clipboardCopy(string prefix);
	void clipboardPaste(string prefix);

	//Actions (with Public Access)
	//--------------------------------------------
	string resizeCommand;
	string focusCommand;
	string unfocusCommand;

	string successCommand;
	string failCommand;

	//AutoSize
	int autoWidthMode;
	int autoHeightMode;

	//Internal Used Settings (with Public Access)
	//--------------------------------------------
	//ModuleHook
	bool hasHookedModules;
	StringList moduleHook;
	
	//Scroll AddOns
	bool bFrozenScroll;
	int iScrollBackup;
	
	//Docking
	StringList dockedToList;
	LabelList hasDockedLabelList;
	
	//Type: 0 left, 1 top, 2 right, 3 bottom, topleft 4, topright 5, bottomleft 6, bottomright 7
	int dockedtoSide;

	bool FindDockedLabel(string strFindLabelName);
	
	//TextEditBox
	WNDPROC wpOrigEditProc;	
	
	//GhostLabels
	GhostLabelList ghostlabelList;

private:

	void CleanupPointers();

	bool bParentDestroyNotificationReceived;

	Label_RegionList labelGhostRegions;

	HCURSOR cursor;
	HCURSOR textcursor;
	HCURSOR handcursor;
	string cursorPath;
	string textcursorPath;

	//Label Group
	string strLabelGroup;

	HWND hWnd;
	HWND hWndOwner;
	HINSTANCE hInstance;
	HWND box;

	HWND hWndTextBox;

	HFONT textboxhFont;
	string textboxXOrg;
	string textboxYOrg;
	string textboxWidthOrg;
	int textboxHeight;
	int tebleftBorder;
	int tebrightBorder;
	int tebtopBorder;
	int tebbottomBorder;
	bool bTextboxVisible;
	COLORREF textboxTextColor;

	RECT TextEditBoxSizeBorder;
	RECT TextEditBoxSize;

	string strEnterAction;
	void EnterAction_Processing(string strInput);

	//Tooltip
	PaintTooltip *pTooltip;

	//Texture 
	PaintTexture *normalbackground;
	PaintTexture *hoverbackground;
	PaintTexture *pressedbackground;

	PaintTexture *normalanim;
	PaintTexture *hoveranim;

	PaintTexture *texteditboxbackground;

	//Font
	PaintHTMLText *normalfont;
	PaintHTMLText *hoverfont;
	PaintHTMLText *pressedfont;
	
	bool bLineBreak;
	bool bHoverLineBreak;
	bool bPressedLineBreak;

	bool bNoMoveCursor;

	bool bNormalWindowzOrder;

	string strBehindWindow;
	HWND hWndBehindWindow;

	bool bAlwaysOnTop;
	bool bStartHidden;

	int actualx;
	int actualy;
	int actualwidth;
	int actualheight;

	bool bScrollingPaused;
	int iScrollPadLength;
	int iScrollInterval;
	int iScrollSpeed;
	int iScroll;
	int iScrollPosition;
	int iScrollLimit;

	int iScrollPause;
	int iScrollActualStep;
	int iScrollSteps;
	int iScrollStepsInitial;

	bool bRestartScrollRepaint;
	bool bSkipScrollRepaint;

	//AutoSize
	int autoMinWidth;
	int autoMinHeight;
	int autoMaxWidth;
	int autoMaxHeight;
	int startCenterX;
	int startCenterY;

	int autoAnimSteps;
	int autoAnimDelay;

	int oldWidth;
	int oldHeight;

	//Ghosted 
	bool bGhosted;

	bool bTrueTransparency;
	bool bTrackingIgnoreTrueTransparency;

	HRGN rgnAlphaMap;

	int alphaTransparency;
	bool alphaMap;
	bool fakealphaMap;
	int alphaFade;
	int alphaFadeTransparencyBackup;
	int alphaFadeBangStep;
	int alphaFadeBangEndValue;

	int updateInterval;
	bool alwaysUpdateContent;

	bool dynamicText;

	string strExtraText;
	string strExtraHoverText;

	vector<string> originalText;
	string originalTextBackup;
	int current;
	int currentBackup;
	bool noTextSplit;
	string text;
	vector<string> hrefconfig;
	bool noEscape;

	bool dynamicTooltip;
	HWND hWndTooltip;
	string strOriginalTooltip;
	string strTooltip;

	bool focusOnEnter;
	bool skipButton;

	bool autoHideIsRunning;

	//Actions
	string leftClickCommand;
	string leftClickDownCommand;
	string leftClickUpCommand;
	string leftDoubleClickCommand;

	string middleClickCommand;
	string middleClickDownCommand;
	string middleClickUpCommand;
	string middleDoubleClickCommand;

	string rightClickCommand;
	string rightClickDownCommand;
	string rightClickUpCommand;
	string rightDoubleClickCommand;

	string X1ClickCommand;
	string X1ClickDownCommand;
	string X1ClickUpCommand;
	string X1DoubleClickCommand;

	string X2ClickCommand;
	string X2ClickDownCommand;
	string X2ClickUpCommand;
	string X2DoubleClickCommand;

	ModClickList leftModClickList;
	ModClickList leftModClickDownList;
	ModClickList leftModClickUpList;
	ModClickList leftModDoubleClickList;

	ModClickList middleModClickList;
	ModClickList middleModClickDownList;
	ModClickList middleModClickUpList;
	ModClickList middleModDoubleClickList;

	ModClickList rightModClickList;
	ModClickList rightModClickDownList;
	ModClickList rightModClickUpList;
	ModClickList rightModDoubleClickList;

	ModClickList X1ModClickList;
	ModClickList X1ModClickDownList;
	ModClickList X1ModClickUpList;
	ModClickList X1ModDoubleClickList;

	ModClickList X2ModClickList;
	ModClickList X2ModClickDownList;
	ModClickList X2ModClickUpList;
	ModClickList X2ModDoubleClickList;

	void FillModClickList(StringList inputList, ModClickList &outputList);
	bool CheckModClickList(ModClickList inputList);

	string wheelDownCommand;
	string wheelUpCommand;
	string enterCommand;
	string leaveCommand;
	string dropCommand;
	string textChangeCommand;
	string moveCommand;
	string hideCommand;
	string showCommand;
	string hoverCommand;

	string gainfocusCommand;
	string loosefocusCommand;

	string animloopstartCommand;
	string animloopstopCommand;

	bool bFocusActive;

	string keyDownCommand;
	string keyUpCommand;

	int hoverTime;

	bool fireOnMoveDuringMove;
	bool fireOnResizeDuringResize;

	string strSnapTo;
	int iSnapDistance;
	int iDockOnSnapTo;
	StringList snapToList;

	bool bMoveable;
	int iMoveKey;
	int iMoveButton;

	bool bUseCustomMoveArea;
	RECT rectCustomMoveArea;

	bool bUseResizeBorder;
	int iResizeBorderLeft;
	int iResizeBorderRight;
	int iResizeBorderTop;
	int iResizeBorderBottom;
	bool bResizeBorderResize;
	POINT ptResizeStepSize;

	Label_RegionList labelLeftClickRegions;
	Label_RegionList labelLeftDoubleClickRegions;
	
	Label_RegionList labelMiddleClickRegions;
	Label_RegionList labelMiddleDoubleClickRegions;

	Label_RegionList labelRightClickRegions;
	Label_RegionList labelRightDoubleClickRegions;

	Label_RegionList labelX1ClickRegions;
	Label_RegionList labelX1DoubleClickRegions;

	Label_RegionList labelX2ClickRegions;
	Label_RegionList labelX2DoubleClickRegions;

	Label_RegionList labelTooltipRegions;

	void FillLabel_Regions(StringList inputList, Label_RegionList &outputList, bool tooltipregion = false);
	void UpdateLabel_Regions(Label_RegionList inputList);

	//Animations
	int iFrameCounter;
	int iLoopCounter;
	int iHoverLoopCounter;

	int imageLoop;
	int imageFrameDelay;
	int imageFrameCount;
	int hoverimageLoop;
	int hoverimageFrameDelay;
	int hoverimageFrameCount;

	int alphaFadeStep;
	int alphaFadeDelay;

	//Internal Used Settings
	//------------------------
	//System
	string strName;
	bool bLabelInitialized;

	bool bVisible;

	//Alpha Features
	bool bLayeredFlag;
	int iUpdateAlpha;

	//Moving Features
	POINT ptSavedMovePoint;
	bool bHasMoved;
	bool bMoveModifierKeyPressed;
	int iMoveButtonPressed;

	int iMoveRestriction;

	//Text Features
	vector<string> listLabelHREFClickRegions;
	bool bMouseInsideHREF;

	//Mouse Features
	bool bMousePressed;
	bool bMouseInside;
	bool bMouseInside2;

	void MouseTrack();

	//Image (Hover/Pressed/Animation) Features
	bool hoverActive;
	bool pressedActive;

	bool useHover;
	bool usePressed;

	bool normalAnim;
	bool hoverAnim;

	//Docking
	POINT oldPositionBackup;
	POINT oldSizeBackup;

	string strDockSenderName;

	//HRGN Overlay Tracking
	void UpdateOverlayHRGN(int width, int height, PaintTexture *pTexture, HRGN &rgnResult);

	bool bHasOverlayConditional;

protected:

	HBRUSH hTextBGBrush;
	HBRUSH GetBGBrush();

	HBITMAP oldbackupBMP;
	HDC backupDC;
	HBITMAP backupBitmap;

	HBITMAP oldscrollBMP;
	HDC scrollDC;
	HBITMAP scrollBitmap;

	void onTemplateButtonDown(int x, int y, int actualMouseClick, Label_RegionList &labelClickRegions, string clickDownCommand);
	void onTemplateButtonUp(int x, int y, int actualMouseClick, Label_RegionList &labelClickRegions, string clickCommand, string clickUpCommand);

	void AutoSize();

	void onTextPaint(HDC &desthdc, int width, int height);
	
	int IsCloseTo(POINT &pos, const POINT size, int x, int y, const int cx, const int cy, const int span, bool skipX = false, bool skipY = false);

	//Alpha (Normal and AlphaMap)
	void UpdateWindowTransparency(int newAlphaTransparency);
	
	void RelayMouseMessageToBox(UINT message, WPARAM wParam, LPARAM lParam);

	virtual void onLButtonDblClk(int x, int y);
	virtual void onLButtonDown(int x, int y);
	virtual void onLButtonUp(int x, int y);
	virtual void onMButtonDblClk(int x, int y);
	virtual void onMButtonDown(int x, int y);
	virtual void onMButtonUp(int x, int y);
	virtual void onRButtonDblClk(int x, int y);
	virtual void onRButtonDown(int x, int y);
	virtual void onRButtonUp(int x, int y);
	virtual void onX1ButtonDown(int x, int y);
	virtual void onX1ButtonUp(int x, int y);
	virtual void onX1ButtonDblClk(int x, int y);
	virtual void onX2ButtonDown(int x, int y);
	virtual void onX2ButtonUp(int x, int y);
	virtual void onX2ButtonDblClk(int x, int y);
	virtual void onWheelDown(int x, int y);
	virtual void onWheelUp(int x, int y);
	virtual void onMouseEnter();
	virtual void onMouseLeave();
	virtual void onMouseMove(int x, int y);
	virtual void onPaint(HDC hDC, RECT *specialRect = NULL);
	virtual void onWindowPosChanging(WINDOWPOS *windowPos);
	virtual void onMove(int x, int y);
	virtual void onSize(int width, int height);
	virtual void onTimer(int timerID);

	virtual LRESULT NcHitTest(int x, int y);

	virtual bool onWindowMessage(UINT message, WPARAM wParam, LPARAM lParam, LRESULT &lResult);

public:

	static LRESULT WINAPI windowProcedure(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

};

#endif
